from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from .colorize import Color
from .get_parent_stack_id import get_parent_stack_id

if TYPE_CHECKING:
    from .grammar_graph_types import PrintOpts
    from .graph_node import GraphNode
    from .graph_pointer import GraphPointer

from .type_guards import is_range, is_rule_char, is_rule_ref


def print_graph_pointer(pointer: GraphPointer) -> Callable[[PrintOpts], str]:
    def print_graph_pointer_inner(opts: PrintOpts) -> str:
        col = opts["colorize"]
        return col(
            f"*{get_parent_stack_id(pointer, col)}",
            Color.RED,
        )

    return print_graph_pointer_inner


def print_graph_node(node: GraphNode) -> Callable[[PrintOpts], str]:
    def print_graph_node_inner(opts: PrintOpts) -> str:
        pointers = opts.get("pointers", set())
        col = opts["colorize"]
        show_position = opts.get("show_position", False)
        rule = node.rule
        parts: list[str] = []
        if show_position:
            parts.extend(
                [col("{", Color.BLUE), col(node.id, Color.GRAY), col("}", Color.BLUE)],
            )

        if is_rule_char(rule):
            parts.extend(
                [
                    col("[", Color.GRAY),
                    col(
                        "".join(
                            [
                                (
                                    "".join(
                                        [col(get_char(val), Color.YELLOW) for val in v],
                                    )
                                    if is_range(v)
                                    else get_char(v)
                                )
                                for v in rule.value
                            ],
                        ),
                        Color.YELLOW,
                    ),
                    col("]", Color.GRAY),
                ],
            )

        elif is_rule_ref(rule):
            parts.extend(
                [
                    col("Ref(", Color.GRAY),
                    col(rule.value, Color.GREEN),
                    col(")", Color.GRAY),
                ],
            )

        else:
            parts.append(col(rule.type, Color.YELLOW))

        if pointers:
            for pointer in pointers:
                pointer_parts: list[str] = []
                if pointer.node == node:
                    pointer_parts.append(pointer.print(opts))

                if len(pointer_parts) > 0:
                    parts.extend(
                        [
                            col("[", Color.GRAY),
                            col("".join(pointer_parts), Color.YELLOW),
                            col("]", Color.GRAY),
                        ],
                    )

        parts_to_return = [
            "".join(parts),
        ]
        if node.next:
            parts_to_return.append(node.next.print(opts))

        return col("-> ", Color.GRAY).join(parts_to_return)

    return print_graph_node_inner


def get_char(char_code: int) -> str:
    char = chr(char_code)
    if char == "\n":
        return "\\n"
    return char



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from .grammar_graph_types import PrintOpts, UnresolvedRule
from .graph_node import GraphNode
from .print import print_graph_pointer
from .type_guards import (
    is_graph_pointer_rule_char,
    is_graph_pointer_rule_char_exclude,
    is_graph_pointer_rule_end,
    is_graph_pointer_rule_ref,
    is_rule_end,
)

if TYPE_CHECKING:
    from .graph_node import GraphNode

T = TypeVar("T", bound=UnresolvedRule)

GraphPointerKey = str


class GraphPointer(Generic[T]):
    node: GraphNode[T]
    parent: GraphPointer | None = None
    id: str
    __valid__: bool | None = None

    def __init__(self, node: GraphNode[T], parent: GraphPointer | None = None):
        if node is None:
            raise ValueError("Node is undefined")
        self.node = node
        self.parent = parent
        self.id = f"{parent.id}-{node.id}" if parent else node.id

    @property
    def rule(self) -> T:
        return self.node.rule

    @property
    def valid(self) -> bool | None:
        return self.__valid__

    @valid.setter
    def valid(self, valid: bool) -> None:
        self.__valid__ = valid

    def print(self, opts: PrintOpts):
        return print_graph_pointer(self)(opts)

    def resolve(self, resolved=False):
        """
        Resolve the graph pointer.

        Args:
            resolved (bool, optional): Whether the pointer has already been resolved. Defaults to False.

        Yields:
            ResolvedGraphPointer: The resolved graph pointer.
        """
        if is_graph_pointer_rule_ref(self):
            if resolved:
                if not self.node.next:
                    raise ValueError(f"No next node: {self.node}")
                yield from GraphPointer(self.node.next, self.parent).resolve()
            else:
                for node in self.node.rule.nodes:
                    yield from GraphPointer(node, self).resolve()
        elif is_graph_pointer_rule_end(self):
            if not self.parent:
                yield self
            else:
                yield from self.parent.resolve(True)
        elif is_graph_pointer_rule_char(self) or is_graph_pointer_rule_char_exclude(
            self,
        ):
            yield self
        else:
            raise ValueError(f"Unknown rule: {self.node.rule}")

    def fetch_next(self):
        """
        Fetch the next resolved graph pointer.

        Yields:
            ResolvedGraphPointer: The next resolved graph pointer.
        """
        # If this pointer is invalid, then we don't return any new pointers
        if not self.valid:
            return

        # If this pointer is an end node, we return the parent's next node.
        # If no parent exists, we return nothing, since it's the end of the line.
        if is_rule_end(self.node.rule):
            if self.parent:
                yield from self.parent.fetch_next()
        else:
            if not self.node.next:
                raise ValueError(f"No next node: {self.node}")
            pointer = GraphPointer(self.node.next, self.parent)
            yield from pointer.resolve()



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from .graph_node import GraphNode, GraphNodeMeta, PrintOpts
from .rule_ref import RuleRef

if TYPE_CHECKING:
    pass


@pytest.fixture(autouse=True)
def mock_print_graph_node():
    with patch(
        "gbnf.grammar_graph.graph_node.print_graph_node",
    ) as mock_print_graph_node:

        def mock_print_graph_node_inner(_opts: PrintOpts) -> str:
            return "mocked_response"

        mock_print_graph_node.return_value = mock_print_graph_node_inner
        yield mock_print_graph_node


def describe_graph_node():
    meta: GraphNodeMeta = {"stackId": 1, "pathId": 2, "stepId": 3}
    rule = RuleRef(42)

    def test_construct_with_rule_and_meta():
        node = GraphNode(rule, meta)
        assert node.rule == rule
        assert node.meta == meta

    def test_throw_if_meta_is_undefined():
        with pytest.raises(ValueError, match="Meta is undefined"):
            GraphNode(rule, None)

    def test_correctly_calculate_and_cache_its_id():
        node = GraphNode(rule, meta)
        assert node.id == "1,2,3"
        node.meta = {"stackId": 4, "pathId": 5, "stepId": 6}
        assert node.id == "1,2,3"

    def test_delegate_print_to_the_print_graph_node_function(mock_print_graph_node):
        node = GraphNode(rule, meta)
        opts: PrintOpts = {"colorize": lambda x, _y: str(x), "show_position": False}
        node.print(opts)

        mock_print_graph_node.assert_called_with(node)

    def test_handle_next_node_linkage():
        nextNode = GraphNode(RuleRef(43), meta)
        node = GraphNode(rule, meta, nextNode)
        assert node.next == nextNode



from time import perf_counter

from .errors import GrammarParseError
from .gbnf_types import InternalRuleType
from .is_word_char import is_word_char
from .parse_char import parse_char
from .parse_name import parse_name
from .parse_space import parse_space


class RulesBuilder:
    def __init__(self, src, limit=1000):
        self.pos = 0
        self.symbol_ids = {}
        self.rules = []
        self.src = src
        self.start = perf_counter()
        self.time_limit = limit
        self.parse(src)

    def parse(self, src):
        self.pos = parse_space(src, 0, True)
        while self.pos < len(src):
            self.parse_rule(src)

        # Validate the state to ensure that all rules are defined
        for rule in self.rules:
            for elem in rule:
                if elem["type"] == InternalRuleType.RULE_REF:
                    if (
                        elem["value"] >= len(self.rules)
                        or not self.rules[elem["value"]]
                    ):
                        for key, value in self.symbol_ids.items():
                            if value == elem["value"]:
                                raise GrammarParseError(
                                    src,
                                    self.pos,
                                    f"Undefined rule identifier '{key}'",
                                )

    def parse_rule(self, src):
        name = parse_name(src, self.pos)
        self.pos = parse_space(src, self.pos + len(name), False)
        rule_id = self.get_symbol_id(name, len(name))

        self.pos = parse_space(src, self.pos, True)
        if not (
            self.pos + 2 < len(src)  # Ensure the position + 2 is within bounds
            and src[self.pos] == ":"
            and src[self.pos + 1] == ":"
            and src[self.pos + 2] == "="
        ):
            raise GrammarParseError(src, self.pos, f"Expecting ::= at {self.pos}")
        self.pos += 3
        self.pos = parse_space(src, self.pos, True)

        self.parse_alternates(name, rule_id)

        # Check if self.pos is within the bounds of src before checking for a carriage return
        if self.pos < len(src) and src[self.pos] == "\r":
            self.pos += 2 if src[self.pos + 1] == "\n" else 1
        elif self.pos < len(src) and src[self.pos] == "\n":
            self.pos += 1
        elif self.pos < len(src) and src[self.pos]:
            raise GrammarParseError(
                src,
                self.pos,
                f"Expecting newline or end at {self.pos}",
            )
        self.pos = parse_space(src, self.pos, True)

    def get_symbol_id(self, src, length):
        key = src[:length]
        if key not in self.symbol_ids:
            self.symbol_ids[key] = len(self.symbol_ids)
        return self.symbol_ids[key]

    def generate_symbol_id(self, base_name):
        next_id = len(self.symbol_ids)
        new_name = f"{base_name}_{next_id}"
        self.symbol_ids[new_name] = next_id
        return next_id

    def add_rule(self, rule_id, rule):
        while len(self.rules) <= rule_id:
            self.rules.append([])
        self.rules[rule_id] = rule

    def check_duration(self):
        if perf_counter() - self.start > self.time_limit:
            raise GrammarParseError(
                self.src,
                self.pos,
                f"Duration of {self.time_limit} exceeded",
            )

    def parse_sequence(self, rule_name, out_elements, depth=0):
        is_nested = depth != 0
        src = self.src
        last_sym_start = len(out_elements)

        while self.pos < len(src):
            if src[self.pos] == '"':
                self.pos += 1
                last_sym_start = len(out_elements)
                while src[self.pos] != '"':
                    self.check_duration()
                    value, inc_pos = parse_char(src, self.pos)
                    out_elements.append(
                        {"type": InternalRuleType.CHAR, "value": [value]},
                    )
                    self.pos += inc_pos
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif src[self.pos] == "[":
                self.pos += 1
                start_type = InternalRuleType.CHAR
                if src[self.pos] == "^":
                    self.pos += 1
                    start_type = InternalRuleType.CHAR_NOT
                last_sym_start = len(out_elements)
                while src[self.pos] != "]":
                    self.check_duration()
                    type_ = (
                        InternalRuleType.CHAR_ALT
                        if last_sym_start < len(out_elements)
                        else start_type
                    )
                    startchar_value, inc_pos = parse_char(src, self.pos)
                    self.pos += inc_pos
                    out_elements.append(
                        {
                            "type": type_,
                            "value": (
                                [startchar_value]
                                if type_
                                in (InternalRuleType.CHAR, InternalRuleType.CHAR_NOT)
                                else startchar_value
                            ),
                        },
                    )

                    if src[self.pos] == "-" and src[self.pos + 1] != "]":
                        self.pos += 1
                        endchar_value, inc_pos = parse_char(src, self.pos)
                        out_elements.append(
                            {
                                "type": InternalRuleType.CHAR_RNG_UPPER,
                                "value": endchar_value,
                            },
                        )
                        self.pos += inc_pos
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif is_word_char(src[self.pos]):
                name = parse_name(src, self.pos)
                ref_rule_id = self.get_symbol_id(name, len(name))
                self.pos += len(name)
                self.pos = parse_space(src, self.pos, is_nested)

                last_sym_start = len(out_elements)
                out_elements.append(
                    {"type": InternalRuleType.RULE_REF, "value": ref_rule_id},
                )
            elif src[self.pos] == "(":
                self.pos = parse_space(src, self.pos + 1, True)
                sub_rule_id = self.generate_symbol_id(rule_name)
                self.parse_alternates(rule_name, sub_rule_id, depth + 1)
                last_sym_start = len(out_elements)
                out_elements.append(
                    {"type": InternalRuleType.RULE_REF, "value": sub_rule_id},
                )
                if src[self.pos] != ")":
                    raise GrammarParseError(
                        src,
                        self.pos,
                        f"Expecting ')' at {self.pos}",
                    )
                self.pos = parse_space(src, self.pos + 1, is_nested)
            elif src[self.pos] in "*+?":
                if last_sym_start == len(out_elements):
                    raise GrammarParseError(
                        src,
                        self.pos,
                        f"Expecting preceding item to */+/? at {self.pos}",
                    )
                sub_rule_id = self.generate_symbol_id(rule_name)
                sub_rule = out_elements[last_sym_start:]
                if src[self.pos] in "*+":
                    sub_rule.append(
                        {"type": InternalRuleType.RULE_REF, "value": sub_rule_id},
                    )
                sub_rule.append({"type": InternalRuleType.ALT})
                if src[self.pos] == "+":
                    sub_rule.extend(out_elements[last_sym_start:])
                sub_rule.append({"type": InternalRuleType.END})
                self.add_rule(sub_rule_id, sub_rule)
                out_elements[last_sym_start:] = [
                    {"type": InternalRuleType.RULE_REF, "value": sub_rule_id},
                ]
                self.pos = parse_space(src, self.pos + 1, is_nested)
            else:
                break

    def parse_alternates(self, rule_name, rule_id, depth=0):
        src = self.src
        rule = []
        self.parse_sequence(rule_name, rule, depth)
        # Ensure that self.pos is within bounds before checking src[self.pos]
        while self.pos < len(src) and src[self.pos] == "|":
            self.check_duration()
            rule.append({"type": InternalRuleType.ALT})
            self.pos = parse_space(src, self.pos + 1, True)
            self.parse_sequence(rule_name, rule, depth)
        rule.append({"type": InternalRuleType.END})
        self.add_rule(rule_id, rule)



import pytest

from .gbnf_types import InternalRuleType
from .rules_builder import RulesBuilder

test_cases = [
    (
        "single-string",
        'root ::= "foo"',
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("o")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("o")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "quote character",
        r'root ::= "\""',
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord('"')]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "two-lines-referencing-expression",
        '''root ::= foo
        foo ::= "bar"''',
        (
            [("root", 0), ("foo", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("b")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("r")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "expression with dash",
        '''root ::= foo-bar
        foo-bar ::= "bar"''',
        (
            [("root", 0), ("foo-bar", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("b")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR, "value": [ord("r")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "simple-grammar",
        """
        root  ::= (expr "=" term "\n")+
        expr  ::= term ([-+*/] term)*
        term  ::= [0-9]+
        """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("term", 3),
                ("root_4", 4),
                ("expr_5", 5),
                ("expr_6", 6),
                ("term_7", 7),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [61]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 43},
                    {"type": InternalRuleType.CHAR_ALT, "value": 42},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "longer-grammar",
        """
        root  ::= (expr "=" ws term "\n")+
        expr  ::= term ([-+*/] term)*
        term  ::= ident | num | "(" ws expr ")" ws
        ident ::= [a-z] [a-z0-9_]* ws
        num   ::= [0-9]+ ws
        ws    ::= [ \t\n]*
        """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("ws", 3),
                ("term", 4),
                ("root_5", 5),
                ("expr_6", 6),
                ("expr_7", 7),
                ("ident", 8),
                ("num", 9),
                ("ident_10", 10),
                ("num_11", 11),
                ("ws_12", 12),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [61]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 12},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 8},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 9},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [40]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [41]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 43},
                    {"type": InternalRuleType.CHAR_ALT, "value": 42},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 11},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.CHAR_ALT, "value": 48},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 95},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 11},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [32]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 9},
                    {"type": InternalRuleType.CHAR_ALT, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 12},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "character unicode grammar",
        "root  ::= [ぁ-ゟ]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("ぁ")]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("ゟ")},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "character alts grammar",
        "root  ::= [az]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("z")},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "character range grammar",
        "root  ::= [a-zA-Z0-9]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("z")},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("A")},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("Z")},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("0")},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("9")},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "character range grammar with dash at end",
        "root  ::= [a-z-]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("z")},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("-")},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "grouping",
        'root  ::= "f" ("b" | "a")',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("b")]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "optional",
        'root  ::= "f"?',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "repeating",
        'root  ::= "f"*',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "repeating-at-least-once",
        'root  ::= "f"+',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "character range with optional repeating",
        "root  ::= [a-zA-Z0-9]*",
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("z")},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("A")},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("Z")},
                    {"type": InternalRuleType.CHAR_ALT, "value": ord("0")},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("9")},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "grouping-repeating",
        'root  ::= "f" ("b" | "a")*',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("f")]},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [ord("b")]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [ord("a")]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "negation",
        r"root ::= [^\n]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [ord("\n")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "negation of range",
        "root ::= [^0-9]",
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [ord("0")]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": ord("9")},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "negation with after",
        r'root ::= [^\n]+ "\n"',
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.CHAR, "value": [ord("\n")]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [ord("\n")]},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR_NOT, "value": [ord("\n")]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "longer negation",
        r'root ::= "\"" ( [^"abcdefgh])* ',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_ALT, "value": 98},
                    {"type": InternalRuleType.CHAR_ALT, "value": 99},
                    {"type": InternalRuleType.CHAR_ALT, "value": 100},
                    {"type": InternalRuleType.CHAR_ALT, "value": 101},
                    {"type": InternalRuleType.CHAR_ALT, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 103},
                    {"type": InternalRuleType.CHAR_ALT, "value": 104},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "longer negation with a range",
        r'root ::= "\"" ( [^"abcdefghA-Z])* ',
        (
            [("root", 0), ("root_1", 1), ("root_2", 2)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_ALT, "value": 98},
                    {"type": InternalRuleType.CHAR_ALT, "value": 99},
                    {"type": InternalRuleType.CHAR_ALT, "value": 100},
                    {"type": InternalRuleType.CHAR_ALT, "value": 101},
                    {"type": InternalRuleType.CHAR_ALT, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 103},
                    {"type": InternalRuleType.CHAR_ALT, "value": 104},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 90},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    *[
        (
            key,
            f'root ::= "{escaped_char}"',
            (
                [("root", 0)],
                [
                    [
                        {"type": InternalRuleType.CHAR, "value": [actual_char]},
                        {"type": InternalRuleType.END},
                    ],
                ],
            ),
        )
        for key, escaped_char, actual_char in [
            ("escaped 8-bit unicode char", r"\x2A", ord("\x2A")),
            ("escaped 16-bit unicode char", r"\u006F", ord("\u006F")),
            ("escaped 32-bit unicode char", r"\U0001F4A9", 128169),
            ("escaped tab char", r"\t", ord("\t")),
            ("escaped new line char", r"\n", ord("\n")),
            ("escaped \r char", r"\r", ord("\r")),
            ("escaped quote char", r"\"", ord('"')),
            ("escaped [ char", r"\[", ord("[")),
            ("escaped ] char", r"\]", ord("]")),
            ("escaped \\ char", r"\\", ord("\\")),
        ]
    ],
    (
        "simple arithmetic",
        r"""
        root ::= (expr "=" term "\n")+
        expr ::= term ([-+*/] term)*
        term ::= [0-9]+
        """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("term", 3),
                ("root_4", 4),
                ("expr_5", 5),
                ("expr_6", 6),
                ("term_7", 7),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [61]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 43},
                    {"type": InternalRuleType.CHAR_ALT, "value": 42},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "ranges with chars",
        """
        root ::= [a-z0-9_]*
        """,
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.CHAR_ALT, "value": 48},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 95},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "nested ranges with chars",
        """
        root ::= [a-z] [a-z0-9_]*
        """,
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.CHAR_ALT, "value": 48},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 95},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "expression with nested range with chars",
        r"""
        root ::= ident
        ident ::= [a-z] [a-z0-9_]* ws
        ws ::= [ \t\n]*
        """,
        (
            [("root", 0), ("ident", 1), ("ident_2", 2), ("ws", 3), ("ws_4", 4)],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.CHAR_ALT, "value": 48},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 95},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [32]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 9},
                    {"type": InternalRuleType.CHAR_ALT, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "lots of escapes",
        r'root ::= "\x2A" "\u006F" "\U0001F4A9" "\t" "\n" "\r" "\"" "\[" "\]" "\\"',
        (
            [("root", 0)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [42]},  # \x2A
                    {"type": InternalRuleType.CHAR, "value": [111]},  # \u006F
                    {"type": InternalRuleType.CHAR, "value": [128169]},  # \U0001F4A9
                    {"type": InternalRuleType.CHAR, "value": [9]},  # \t
                    {"type": InternalRuleType.CHAR, "value": [10]},  # \n
                    {"type": InternalRuleType.CHAR, "value": [13]},  # \r
                    {"type": InternalRuleType.CHAR, "value": [34]},  # \"
                    {"type": InternalRuleType.CHAR, "value": [91]},  # \[
                    {"type": InternalRuleType.CHAR, "value": [93]},  # \]
                    {"type": InternalRuleType.CHAR, "value": [92]},  # \\
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "lots of escape and alternate escapes",
        r"""root ::= "\x2A" "\u006F" "\U0001F4A9" "\t" "\n" "\r" "\"" "\[" "\]" "\\" (
            "\x2A" | "\u006F" | "\U0001F4A9" | "\t" | "\n" | "\r" | "\"" | "\[" | "\]"  | "\\" )""",
        (
            [("root", 0), ("root_1", 1)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [42]},
                    {"type": InternalRuleType.CHAR, "value": [111]},
                    {"type": InternalRuleType.CHAR, "value": [128169]},
                    {"type": InternalRuleType.CHAR, "value": [9]},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.CHAR, "value": [13]},
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.CHAR, "value": [91]},
                    {"type": InternalRuleType.CHAR, "value": [93]},
                    {"type": InternalRuleType.CHAR, "value": [92]},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [42]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [111]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [128169]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [9]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [13]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [91]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [93]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [92]},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "arithmetic",
        r"""
        root  ::= (expr "=" ws term "\n")+
        expr  ::= term ([-+*/] term)*
        term  ::= ident | num | "(" ws expr ")" ws
        ident ::= [a-z] [a-z0-9_]* ws
        num   ::= [0-9]+ ws
        ws    ::= [ \t\n]*
        """,
        (
            [
                ("root", 0),
                ("root_1", 1),
                ("expr", 2),
                ("ws", 3),
                ("term", 4),
                ("root_5", 5),
                ("expr_6", 6),
                ("expr_7", 7),
                ("ident", 8),
                ("num", 9),
                ("ident_10", 10),
                ("num_11", 11),
                ("ws_12", 12),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [61]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.CHAR, "value": [10]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 12},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 8},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 9},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [40]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.CHAR, "value": [41]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 43},
                    {"type": InternalRuleType.CHAR_ALT, "value": 42},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 11},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 122},
                    {"type": InternalRuleType.CHAR_ALT, "value": 48},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 95},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 11},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [32]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 9},
                    {"type": InternalRuleType.CHAR_ALT, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 12},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "json.gbnf (string)",
        r"""
        root ::=
        "\"" (
            [^"\\\x7F\x00-\x1F] |
            "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]) # escapes
        )* "\""
        """,
        (
            [("root", 0), ("root_1", 1), ("root_2", 2), ("root_3", 3)],
            [
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 92},
                    {"type": InternalRuleType.CHAR_ALT, "value": 127},
                    {"type": InternalRuleType.CHAR_ALT, "value": 0},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 31},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [92]},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 92},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.CHAR_ALT, "value": 98},
                    {"type": InternalRuleType.CHAR_ALT, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 110},
                    {"type": InternalRuleType.CHAR_ALT, "value": 114},
                    {"type": InternalRuleType.CHAR_ALT, "value": 116},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [117]},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "json.gbnf (full)",
        r"""
        root   ::= object
        value  ::= object | array | string | number | ("true" | "false" | "null") ws
        object ::=
          "{" ws (
                    string ":" ws value
            ("," ws string ":" ws value)*
          )? "}" ws
        array  ::=
          "[" ws (
                    value
            ("," ws value)*
          )? "]" ws
              string ::=
          "\"" (
            [^"\\\x7F\x00-\x1F] |
            "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]) # escapes
          )* "\"" ws
        number ::= ("-"? ([0-9] | [1-9] [0-9]*)) ("." [0-9]+)? ([eE] [-+]? [0-9]+)? ws
        # Optional space: by convention, applied in this grammar after literal chars when allowed
        ws ::= ([ \t\n] ws)?
        """,
        (
            [
                ("root", 0),
                ("object", 1),
                ("value", 2),
                ("array", 3),
                ("string", 4),
                ("number", 5),
                ("value_6", 6),
                ("ws", 7),
                ("object_8", 8),
                ("object_9", 9),
                ("object_10", 10),
                ("object_11", 11),
                ("array_12", 12),
                ("array_13", 13),
                ("array_14", 14),
                ("array_15", 15),
                ("string_16", 16),
                ("string_17", 17),
                ("string_18", 18),
                ("number_19", 19),
                ("number_20", 20),
                ("number_21", 21),
                ("number_22", 22),
                ("number_23", 23),
                ("number_24", 24),
                ("number_25", 25),
                ("number_26", 26),
                ("number_27", 27),
                ("number_28", 28),
                ("number_29", 29),
                ("ws_30", 30),
                ("ws_31", 31),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [123]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 11},
                    {"type": InternalRuleType.CHAR, "value": [125]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [91]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 15},
                    {"type": InternalRuleType.CHAR, "value": [93]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.RULE_REF, "value": 18},
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 19},
                    {"type": InternalRuleType.RULE_REF, "value": 25},
                    {"type": InternalRuleType.RULE_REF, "value": 29},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [116]},
                    {"type": InternalRuleType.CHAR, "value": [114]},
                    {"type": InternalRuleType.CHAR, "value": [117]},
                    {"type": InternalRuleType.CHAR, "value": [101]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [102]},
                    {"type": InternalRuleType.CHAR, "value": [97]},
                    {"type": InternalRuleType.CHAR, "value": [108]},
                    {"type": InternalRuleType.CHAR, "value": [115]},
                    {"type": InternalRuleType.CHAR, "value": [101]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [110]},
                    {"type": InternalRuleType.CHAR, "value": [117]},
                    {"type": InternalRuleType.CHAR, "value": [108]},
                    {"type": InternalRuleType.CHAR, "value": [108]},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 31},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.CHAR, "value": [58]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [44]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.CHAR, "value": [58]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 9},
                    {"type": InternalRuleType.RULE_REF, "value": 10},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 8},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.RULE_REF, "value": 14},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [44]},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 13},
                    {"type": InternalRuleType.RULE_REF, "value": 14},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 12},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR_NOT, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 92},
                    {"type": InternalRuleType.CHAR_ALT, "value": 127},
                    {"type": InternalRuleType.CHAR_ALT, "value": 0},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 31},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [92]},
                    {"type": InternalRuleType.RULE_REF, "value": 17},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [34]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 92},
                    {"type": InternalRuleType.CHAR_ALT, "value": 47},
                    {"type": InternalRuleType.CHAR_ALT, "value": 98},
                    {"type": InternalRuleType.CHAR_ALT, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 110},
                    {"type": InternalRuleType.CHAR_ALT, "value": 114},
                    {"type": InternalRuleType.CHAR_ALT, "value": 116},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [117]},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.CHAR_ALT, "value": 97},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 102},
                    {"type": InternalRuleType.CHAR_ALT, "value": 65},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 70},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 16},
                    {"type": InternalRuleType.RULE_REF, "value": 18},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 20},
                    {"type": InternalRuleType.RULE_REF, "value": 21},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [49]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 22},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 22},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [46]},
                    {"type": InternalRuleType.RULE_REF, "value": 24},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 24},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 23},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [101]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 69},
                    {"type": InternalRuleType.RULE_REF, "value": 27},
                    {"type": InternalRuleType.RULE_REF, "value": 28},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [45]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 43},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.RULE_REF, "value": 28},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.CHAR, "value": [48]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 57},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 26},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [32]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 9},
                    {"type": InternalRuleType.CHAR_ALT, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 30},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
    (
        "japanese",
        r"""
        # A probably incorrect grammar for Japanese
        root        ::= jp-char+ ([ \t\n] jp-char+)*
        jp-char     ::= hiragana | katakana | punctuation | cjk
        hiragana    ::= [ぁ-ゟ]
        katakana    ::= [ァ-ヿ]
        punctuation ::= [、-〾]
        cjk         ::= [一-鿿]
        """,
        (
            [
                ("root", 0),
                ("jp-char", 1),
                ("root_2", 2),
                ("root_3", 3),
                ("root_4", 4),
                ("root_5", 5),
                ("hiragana", 6),
                ("katakana", 7),
                ("punctuation", 8),
                ("cjk", 9),
            ],
            [
                [
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 6},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 7},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 8},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 9},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 2},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [32]},
                    {"type": InternalRuleType.CHAR_ALT, "value": 9},
                    {"type": InternalRuleType.CHAR_ALT, "value": 10},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.RULE_REF, "value": 4},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.RULE_REF, "value": 1},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.RULE_REF, "value": 3},
                    {"type": InternalRuleType.RULE_REF, "value": 5},
                    {"type": InternalRuleType.ALT},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [12353]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 12447},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [12449]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 12543},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [12289]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 12350},
                    {"type": InternalRuleType.END},
                ],
                [
                    {"type": InternalRuleType.CHAR, "value": [19968]},
                    {"type": InternalRuleType.CHAR_RNG_UPPER, "value": 40959},
                    {"type": InternalRuleType.END},
                ],
            ],
        ),
    ),
]


@pytest.mark.parametrize(("key", "grammar", "expected"), test_cases)
def test_grammar_parser(key, grammar, expected):
    symbol_ids_expected, rules_expected = expected
    parsed_grammar = RulesBuilder(grammar.replace("\\n", "\n"))
    # for i, rules in enumerate(parsed_grammar.rules):
    #     for j, rule in enumerate(rules):
    #         # print(i, j, "expected", rules_expected[i][j], "received", rule)
    #         assert rule == rules_expected[i][j]
    assert parsed_grammar.rules == rules_expected
    assert list(parsed_grammar.symbol_ids.items()) == symbol_ids_expected



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from .colorize import Color

if TYPE_CHECKING:
    from .graph_pointer import GraphPointer


def get_parent_stack_id(
    pointer: GraphPointer,
    col: Callable[[str | int, str], str],
) -> str:
    stack_ids: list[str] = []
    parent: GraphPointer | None = pointer.parent
    while parent:
        stack_ids.append(
            f"{parent.node.meta['stackId']},{parent.node.meta['pathId']},{parent.node.meta['stepId']}",
        )
        parent = parent.parent
    arrow = col("<-", Color.GRAY)
    return arrow.join([col(stack_id, Color.RED) for stack_id in stack_ids])



from __future__ import annotations

from unittest.mock import patch

import pytest

from .grammar_graph_types import (
    RuleChar,
    RuleCharExclude,
    RuleEnd,
)
from .graph_node import GraphNode
from .graph_pointer import GraphPointer
from .type_guards import (
    is_graph_pointer_rule_char,
    is_graph_pointer_rule_end,
)


@pytest.fixture(autouse=True)
def mock_is_graph_pointer_rule_ref():
    with patch(
        "gbnf.grammar_graph.graph_pointer.is_graph_pointer_rule_ref", return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_graph_pointer_rule_end():
    with patch(
        "gbnf.grammar_graph.graph_pointer.is_graph_pointer_rule_end", return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_graph_pointer_rule_char():
    with patch(
        "gbnf.grammar_graph.graph_pointer.is_graph_pointer_rule_char",
        return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_graph_pointer_rule_char_exclude():
    with patch(
        "gbnf.grammar_graph.graph_pointer.is_graph_pointer_rule_char_exclude",
        return_value=False,
    ) as mock:
        yield mock


def describe_graph_pointer():
    def test_constructor():
        node = GraphNode(RuleEnd(), {"stackId": 1, "pathId": 2, "stepId": 3})
        pointer = GraphPointer(node)
        assert pointer.node == node
        assert pointer.id == "1,2,3"
        assert pointer.parent is None

    def test_it_raises_error_if_node_is_undefined():
        with pytest.raises(ValueError, match="Node is undefined"):
            GraphPointer(None)

    def test_it_initializes_correctly_with_node_and_parent():
        parent_node = GraphNode(
            RuleChar(value=[97]), {"stackId": 1, "pathId": 2, "stepId": 3},
        )
        child_node = GraphNode(
            RuleChar(value=[98]), {"stackId": 4, "pathId": 5, "stepId": 6},
        )
        parent_pointer = GraphPointer(parent_node)
        child_pointer = GraphPointer(child_node, parent_pointer)
        assert child_pointer.parent == parent_pointer
        assert child_pointer.id == "1,2,3-4,5,6"

    def describe_resolve():
        def test_it_raises_error_on_unknown_rule_types():
            node = GraphNode(
                {type: "UNKNOWN_RULE_TYPE"}, {"stackId": 1, "pathId": 2, "stepId": 3},
            )
            pointer = GraphPointer(node)
            with pytest.raises(ValueError, match="Unknown rule"):
                list(pointer.resolve())

        def test_it_yields_a_char_rule(mock_is_graph_pointer_rule_char):
            mock_is_graph_pointer_rule_char.return_value = True
            rule = RuleChar(value=[97])
            node = GraphNode(rule, {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert list(pointer.resolve()) == [pointer]

        def test_it_yields_a_char_excluded_rule(
            mock_is_graph_pointer_rule_char_exclude,
        ):
            mock_is_graph_pointer_rule_char_exclude.return_value = True
            rule = RuleCharExclude(value=[97])
            node = GraphNode(rule, {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert list(pointer.resolve()) == [pointer]

        def test_it_yields_an_end_rule_without_a_parent(mock_is_graph_pointer_rule_end):
            mock_is_graph_pointer_rule_end.return_value = True
            rule = RuleEnd()
            node = GraphNode(rule, {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert list(pointer.resolve()) == [pointer]

        def test_it_yields_an_end_rule_with_a_parent(mock_is_graph_pointer_rule_end):
            mock_is_graph_pointer_rule_end.return_value = True
            rule = RuleEnd()
            parent_node = GraphNode(rule, {"stackId": 2, "pathId": 1, "stepId": 1})
            parent_pointer = GraphPointer(parent_node)
            child_node = GraphNode(rule, {"stackId": 1, "pathId": 1, "stepId": 1})
            child_pointer = GraphPointer(child_node, parent_pointer)
            assert list(child_pointer.resolve()) == [parent_pointer]

        def test_it_yields_an_end_rule_with_a_grandparent(
            mock_is_graph_pointer_rule_end,
        ):
            mock_is_graph_pointer_rule_end.return_value = True
            rule = RuleEnd()
            grandparent_node = GraphNode(rule, {"stackId": 2, "pathId": 1, "stepId": 1})
            grandparent_pointer = GraphPointer(grandparent_node)
            parent_node = GraphNode(rule, {"stackId": 2, "pathId": 1, "stepId": 1})
            parent_pointer = GraphPointer(parent_node, grandparent_pointer)
            child_node = GraphNode(rule, {"stackId": 1, "pathId": 1, "stepId": 1})
            child_pointer = GraphPointer(child_node, parent_pointer)
            assert list(child_pointer.resolve()) == [grandparent_pointer]

        def test_it_yields_an_end_rule_with_a_parent_that_is_not_an_end_with_a_grandparent(
            mock_is_graph_pointer_rule_end,
            mock_is_graph_pointer_rule_char,
        ):
            mock_is_graph_pointer_rule_end.side_effect = is_graph_pointer_rule_end
            mock_is_graph_pointer_rule_char.side_effect = is_graph_pointer_rule_char
            rule = RuleEnd()
            grandparent_node = GraphNode(rule, {"stackId": 2, "pathId": 1, "stepId": 1})
            grandparent_pointer = GraphPointer(grandparent_node)
            parent_node = GraphNode(
                RuleChar(value=[97]), {"stackId": 2, "pathId": 1, "stepId": 1},
            )
            parent_pointer = GraphPointer(parent_node, grandparent_pointer)
            child_node = GraphNode(rule, {"stackId": 1, "pathId": 1, "stepId": 1})
            child_pointer = GraphPointer(child_node, parent_pointer)
            assert list(child_pointer.resolve()) == [parent_pointer]



from __future__ import annotations

import json

from .colorize import Color
from .get_parent_stack_id import get_parent_stack_id
from .grammar_graph_types import RuleEnd
from .graph_node import GraphNode
from .graph_pointer import GraphPointer


def s(str: str) -> str:
    return json.dumps(str)


def mock_colorize(text: str | int, color: str) -> str:
    return f"[{s(color)}]:{text}"


red = s(Color.RED)
gray = s(Color.GRAY)


def create_mock_pointer(
    stack_id: int,
    path_id: int,
    step_id: int,
    parent: GraphPointer | None = None,
) -> GraphPointer:
    return GraphPointer(
        GraphNode(
            RuleEnd(),
            {
                "stackId": stack_id,
                "pathId": path_id,
                "stepId": step_id,
            },
        ),
        parent,
    )


def describe_get_parent_stack_id():
    def test_returns_an_empty_string_if_no_parents():
        pointer = create_mock_pointer(1, 1, 1)
        result = get_parent_stack_id(pointer, mock_colorize)
        assert result == ""

    def test_returns_a_single_parent_id_colored_correctly():
        parent_pointer = create_mock_pointer(1, 1, 1)
        pointer = create_mock_pointer(2, 2, 2, parent_pointer)
        result = get_parent_stack_id(pointer, mock_colorize)
        assert result == f"[{red}]:1,1,1"

    def test_returns_multiple_parent_ids_separated_by_colored_arrows():
        grandparent_pointer = create_mock_pointer(0, 0, 0)
        parent_pointer = create_mock_pointer(1, 1, 1, grandparent_pointer)
        pointer = create_mock_pointer(2, 2, 2, parent_pointer)
        result = get_parent_stack_id(pointer, mock_colorize)
        assert result == f"[{red}]:1,1,1[{gray}]:<-[{red}]:0,0,0"

    def test_handles_deep_nesting_of_pointers():
        great_grandparent_pointer = create_mock_pointer(0, 0, 0)
        grandparent_pointer = create_mock_pointer(1, 1, 1, great_grandparent_pointer)
        parent_pointer = create_mock_pointer(2, 2, 2, grandparent_pointer)
        pointer = create_mock_pointer(3, 3, 3, parent_pointer)
        result = get_parent_stack_id(pointer, mock_colorize)
        expectedOutput = (
            f"[{red}]:2,2,2[{gray}]:<-[{red}]:1,1,1[{gray}]:<-[{red}]:0,0,0"
        )
        assert result == expectedOutput



from .rules_builder import GrammarParseError, RulesBuilder

ValidInput = str | int | float | list[int | float]


def GBNF(grammar: str, _initial_string: ValidInput = ""):
    #   const grammar = typeof input === 'string' ? input : input.toString();

    rules_builder = RulesBuilder(grammar)
    rules, symbol_ids = rules_builder.rules, rules_builder.symbol_ids
    if len(rules) == 0:
        raise GrammarParseError(grammar, 0, "No rules were found")
    if symbol_ids.get("root") is None:
        raise GrammarParseError(grammar, 0, "Grammar does not contain a 'root' symbol")
    # root_id = symbol_ids.get("root")

    # stacked_rules = list(map(build_rule_stack, rules))
    # graph = Graph(grammar, stacked_rules, root_id)
    # return ParseState(graph, graph.add(initial_string)


#   const stackedRules: UnresolvedRule[][][] = rules.map(buildRuleStack);
#   const graph = new Graph(grammar, stackedRules, rootId);
#   return new ParseState(graph, graph.add(initialString));



import pytest

from .is_point_in_range import is_point_in_range


def describe_is_point_in_range():
    @pytest.mark.parametrize(
        ("point", "range", "expectation"),
        [
            (96, (97, 122), False),
            (97, (97, 122), True),
            (98, (97, 122), True),
            (122, (97, 122), True),
            (123, (97, 122), False),
        ],
    )
    def it_checks_if_point_is_in_range(point, range, expectation):
        assert is_point_in_range(point, range) == expectation



from __future__ import annotations

from unittest.mock import patch

import pytest

from .get_serialized_rule_key import KEY_TRANSLATION, get_serialized_rule_key
from .grammar_graph_types import RuleChar, RuleCharExclude, RuleEnd, RuleRef, RuleType


@pytest.fixture(autouse=True)
def mock_is_rule_end():
    with patch(
        "gbnf.grammar_graph.get_serialized_rule_key.is_rule_end",
        return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_rule_char():
    with patch(
        "gbnf.grammar_graph.get_serialized_rule_key.is_rule_char",
        return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_rule_char_exclude():
    with patch(
        "gbnf.grammar_graph.get_serialized_rule_key.is_rule_char_exclude",
        return_value=False,
    ) as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_is_rule_ref():
    with patch(
        "gbnf.grammar_graph.get_serialized_rule_key.is_rule_ref",
        return_value=False,
    ) as mock:
        yield mock


def describe_get_serialized_rule_key():
    def test_returns_type_for_end_rules(mock_is_rule_end):
        mock_is_rule_end.return_value = True
        rule = RuleEnd()
        assert get_serialized_rule_key(rule) == f"{KEY_TRANSLATION[RuleType.END]}"

    def test_returns_type_and_value_for_character_rules(mock_is_rule_char):
        mock_is_rule_char.return_value = True
        rule = RuleChar(value=[97])
        assert get_serialized_rule_key(rule) == f"{KEY_TRANSLATION[RuleType.CHAR]}-[97]"

    def test_returns_type_and_value_for_character_exclude_rules(
        mock_is_rule_char_exclude,
    ):
        mock_is_rule_char_exclude.return_value = True
        rule = RuleCharExclude(value=[97])
        assert (
            get_serialized_rule_key(rule)
            == f"{KEY_TRANSLATION[RuleType.CHAR_EXCLUDE]}-[97]"
        )

    def test_returns_ref_type_with_value_for_reference_rules(mock_is_rule_ref):
        mock_is_rule_ref.return_value = True
        rule = RuleRef(99)
        assert get_serialized_rule_key(rule) == "3-99"

    def test_throws_error_for_unknown_rule_types(
        mock_is_rule_end,
        mock_is_rule_char,
        mock_is_rule_char_exclude,
        mock_is_rule_ref,
    ):
        mock_is_rule_end.return_value = False
        mock_is_rule_char.return_value = False
        mock_is_rule_char_exclude.return_value = False
        mock_is_rule_ref.return_value = False
        rule = {"type": "UNKNOWN", "value": "something"}
        with pytest.raises(
            ValueError,
            match=r"Unknown rule type: {'type': 'UNKNOWN', 'value': 'something'}",
        ):
            get_serialized_rule_key(rule)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import field
from enum import Enum
from typing import TYPE_CHECKING, Any, TypedDict

from .rule_ref import RuleRef


class PrintOpts(TypedDict):
    pointers: set[GraphPointer]
    colorize: Callable[[str | int, str], str]
    show_position: bool


class RuleType(Enum):
    CHAR = "char"
    CHAR_EXCLUDE = "char_exclude"
    END = "end"


Range = tuple[int, int]


class Rule:
    type: RuleType
    value: Any | None

    @property
    def __dict__(self):
        if self.value is None:
            return {
                "type": self.type,
            }
        return {"type": self.type, "value": self.value}


class RuleChar(Rule):
    type: RuleType = RuleType.CHAR
    value: list[int | Range] = field(default_factory=list)

    def __init__(self, value: list[int | Range]):
        self.value = value


class RuleCharExclude(Rule):
    type: RuleType = RuleType.CHAR_EXCLUDE

    value: list[int | Range] = field(default_factory=list)

    def __init__(self, value: list[int | Range]):
        self.value = value


class RuleEnd(Rule):
    type: RuleType = RuleType.END


UnresolvedRule = RuleChar | RuleCharExclude | RuleRef | RuleEnd


# ValidInput can either be a string, or a number indicating a code point.
# It CANNOT be a number representing a number; a number intended as input (like "8")
# should be passed in as a string.
ValidInput = str | int | list[int]


if TYPE_CHECKING:
    from .graph_pointer import GraphPointer

    # RuleRefs should never be exposed to the end user.
    ResolvedRule = RuleCharExclude | RuleChar | RuleEnd
    ResolvedGraphPointer = GraphPointer[ResolvedRule]
    Pointers = set[ResolvedGraphPointer]



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

from .grammar_graph_types import Rule, RuleType

if TYPE_CHECKING:
    from .grammar_graph_types import Pointers, ResolvedRule
    from .graph import Graph

import json


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, RuleType):
            return o.value
        if isinstance(o, Rule):
            return o
        return super().default(o)


class ParseState:
    __graph__: Graph
    __pointers__: Pointers

    def __init__(self, graph: Graph, pointers: Pointers):
        self.__graph__ = graph
        self.__pointers__ = pointers

    def __call__(self, text: str) -> ParseState:
        return self.add(text)

    def __iter__(self) -> Iterable[ResolvedRule]:
        yield from self.rules()

    def rules(self) -> Iterable[ResolvedRule]:
        rules = set()
        for pointer in self.__pointers__:
            rule = pointer.rule
            key = json.dumps(rule.__dict__, cls=CustomJSONEncoder)
            if key not in rules:
                rules.add(key)
                yield rule

    def add(self, text: str) -> ParseState:
        pointers = self.__graph__.add(text, self.__pointers__)
        return ParseState(self.__graph__, pointers)

    def __add__(self, text: str) -> ParseState:
        return self.add(text)

    @property
    def size(self) -> int:
        return len(list(self.rules()))

    @property
    def grammar(self) -> str:
        return self.__graph__.grammar




from .get_input_as_code_points import get_input_as_code_points


def describe_get_input_as_code_points():
    def it_returns_code_points_for_string():
        assert get_input_as_code_points("abc") == [97, 98, 99]

    def it_returns_code_points_for_number():
        assert get_input_as_code_points(99) == [99]

    def it_returns_code_points_for_array_of_number():
        assert get_input_as_code_points([99, 100, 101]) == [99, 100, 101]




from .grammar_graph_types import ValidInput


def get_code_point(char: str) -> int:
    code_point = ord(char)
    if code_point is None:
        raise ValueError(f"Could not get code point for character: {char}")
    return code_point


def get_input_as_code_points(src: ValidInput) -> list[int]:
    if not isinstance(src, str):
        return list(src) if isinstance(src, list) else [src]

    return [get_code_point(s) for s in src]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from .grammar_graph_types import (
    # Range,
    # RuleType,
    RuleChar,
    RuleCharExclude,
    RuleEnd,
    # UnresolvedRule,
)
from .graph_node import GraphNode
from .graph_pointer import GraphPointer
from .rule_ref import RuleRef

if TYPE_CHECKING:
    # from .graph_pointer import GraphPointer
    pass

from .type_guards import (
    is_graph_pointer_rule_char,
    is_graph_pointer_rule_char_exclude,
    is_graph_pointer_rule_end,
    is_graph_pointer_rule_ref,
    is_range,
    is_rule,
    is_rule_char,
    is_rule_char_exclude,
    is_rule_end,
    # is_rule_type,
    is_rule_ref,
)


def describe_rule_type_guards():
    def describe_is_graph_pointer_rule_ref():
        def test_it_returns_true():
            node = GraphNode(RuleRef(1), {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert is_graph_pointer_rule_ref(pointer)

        def test_it_returns_false():
            node = GraphNode(RuleEnd(), {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert not is_graph_pointer_rule_ref(pointer)

    def describe_is_graph_pointer_rule_end():
        def test_it_returns_false():
            node = GraphNode(RuleRef(1), {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert not is_graph_pointer_rule_end(pointer)

        def test_it_returns_true():
            node = GraphNode(RuleEnd(), {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert is_graph_pointer_rule_end(pointer)

    def describe_is_graph_pointer_rule_char():
        def test_it_returns_true():
            node = GraphNode(
                RuleChar(value=[97]), {"stackId": 1, "pathId": 2, "stepId": 3},
            )
            pointer = GraphPointer(node)
            assert is_graph_pointer_rule_char(pointer)

        def test_it_returns_false():
            node = GraphNode(RuleEnd(), {"stackId": 1, "pathId": 2, "stepId": 3})
            pointer = GraphPointer(node)
            assert not is_graph_pointer_rule_char(pointer)

    def describe_is_graph_pointer_rule_char_exclude():
        def test_it_returns_true():
            node = GraphNode(
                RuleCharExclude(value=[97]),
                {"stackId": 1, "pathId": 2, "stepId": 3},
            )
            pointer = GraphPointer(node)
            assert is_graph_pointer_rule_char_exclude(pointer)

        def test_it_returns_false():
            node = GraphNode(
                RuleChar(value=[97]),
                {"stackId": 1, "pathId": 2, "stepId": 3},
            )
            pointer = GraphPointer(node)
            assert not is_graph_pointer_rule_char_exclude(pointer)

    def describe_is_rule():
        def test_it_returns_false_for_null():
            assert not is_rule(None)

        def test_it_returns_false_for_a_non_rule():
            assert not is_rule({"type": "invalid", "value": []})

        def test_it_returns_true_for_valid_rule_objects():
            ruleChar: RuleChar = RuleChar(value=[65, (66, 67)])
            assert is_rule(ruleChar)

    # def describe_is_rule_type():
    #     def test_it_returns_true_for_valid_rule_types():
    #         assert is_rule_type(RuleType.CHAR)
    #         assert is_rule_type(RuleType.END)

    #     def test_it_returns_false_for_invalid_rule_types():
    #         assert is_rule_type("not_a_rule_type")

    def describe_is_rule_ref():
        def test_it_returns_true_for_valid_rule_refs():
            ruleRef = RuleRef(1)
            assert is_rule_ref(ruleRef)

        def test_it_returns_false_for_invalid_rule_refs():
            assert not is_rule_ref(RuleChar(value=[65]))

    def describe_is_rule_end():
        def test_it_returns_true_for_valid_rule_ends():
            assert is_rule_end(RuleEnd())

        def test_it_returns_false_for_invalid_rule_ends():
            assert not is_rule_end(RuleChar(value=[65]))

    def describe_is_rule_char():
        def test_it_returns_true_for_valid_rule_chars():
            assert is_rule_char(RuleChar(value=[65]))

        def test_it_returns_false_for_invalid_rule_chars():
            assert not is_rule_char(RuleCharExclude(value=[65]))

    def describe_is_rule_char_exclude():
        def test_it_returns_true_for_valid_rule_chars():
            assert is_rule_char_exclude(RuleCharExclude(value=[65]))

        def test_it_returns_false_for_invalid_rule_chars():
            assert not is_rule_char_exclude(RuleChar(value=[65]))

    def describe_is_range():
        def test_it_returns_true_for_valid_ranges():
            assert is_range([1, 10])

        @pytest.mark.parametrize("value", [[1, "10"], [1], 5])
        def test_it_returns_false_for_invalid_ranges(value):
            assert not is_range(value)



class Color:
    BLUE = "\x1b[34m"
    CYAN = "\x1b[36m"
    GREEN = "\x1b[32m"
    RED = "\x1b[31m"
    GRAY = "\x1b[90m"
    YELLOW = "\x1b[33m"


def colorize(text: str | int, color: str) -> str:
    return f"{color}{text}"



import pytest

from .errors import GrammarParseError, InputParseError, build_error_position


@pytest.mark.parametrize(
    ("grammar", "pos", "expected"),
    [
        ('root ::= "foo"', 1, ['root ::= "foo"', " ^"]),
        ('root ::= "foo"', 5, ['root ::= "foo"', "     ^"]),
        ("aa\nbb", 1, ["aa", " ^"]),
        ("aa\nbb", 2, ["aa", "bb", "^"]),
        ("aa\nbb", 2 + 1, ["aa", "bb", " ^"]),
        ("aa\nbb\ncc", 2 + 1, ["aa", "bb", " ^"]),
        ("aa\nbb\ncc", 2 + 2 + 0, ["aa", "bb", "cc", "^"]),
        ("aa\nbb\ncc", 2 + 2 + 1, ["aa", "bb", "cc", " ^"]),
        ("aa\nbb\ncc\ndd", 2 + 2 + 2 + 0, ["bb", "cc", "dd", "^"]),
        ("aa\nbb\ncc\ndd\nee", 2 + 2 + 2 + 2 + 1, ["cc", "dd", "ee", " ^"]),
    ],
)
def test_build_error_position(grammar, pos, expected):
    assert build_error_position(grammar, pos) == expected


def test_build_error_position_for_empty_string():
    assert build_error_position("", 0) == ["No input provided"]


def test_grammar_parse_error():
    grammar = "aa\nbb\ncc\ndd\nee"
    pos = 5
    reason = "reason"
    err = GrammarParseError(grammar, pos, reason)
    expected_message = "Failed to parse grammar: reason\naa\nbb\ncc\n ^"
    assert str(err) == expected_message


def test_input_parse_error():
    input_text = "some input"
    pos = 1
    err = InputParseError(input_text, pos)
    expected_message = "Failed to parse input string:\nsome input\n ^"
    assert str(err) == expected_message


def test_input_parse_error_with_code_point():
    pos = 0
    err = InputParseError(97, pos)
    expected_message = "Failed to parse input string:\na\n^"
    assert str(err) == expected_message


def test_input_parse_error_code_points():
    input_text = [97, 98, 99, 100]  # 'abcd'
    pos = 2
    err = InputParseError(input_text, pos)
    expected_message = "Failed to parse input string:\nabcd\n  ^"
    assert str(err) == expected_message



from .errors import GrammarParseError


def parse_char(src: str, pos: int):
    if pos >= len(src):
        raise GrammarParseError(
            src,
            pos,
            "Unexpected end of grammar input, failed to complete parse",
        )

    if src[pos] == "\\":
        if src[pos + 1] == "x":
            return (int(src[pos + 2 : pos + 4], 16), 4)
        if src[pos + 1] == "u":
            return (int(src[pos + 2 : pos + 6], 16), 6)
        if src[pos + 1] == "U":
            return (int(src[pos + 2 : pos + 10], 16), 10)
        if src[pos + 1] in ["t", "r", "n"]:
            return (ord(eval(f"'\\{src[pos + 1]}'")), 2)
        if src[pos + 1] in ['"', "[", "]", "\\"]:
            return (ord(src[pos + 1]), 2)
        raise GrammarParseError(src, pos, f"Unknown escape at {src[pos]}")

    return (ord(src[pos]), 1)



MAXIMUM_NUMBER_OF_ERROR_LINES_TO_SHOW = 3


def build_error_position(src: str, pos: int) -> list[str]:
    if src == "":
        return [
            "No input provided",
        ]
    lines = src.split("\n")

    line_idx = 0
    while lines[line_idx] and pos > len(lines[line_idx]) - 1:
        pos -= len(lines[line_idx])
        line_idx += 1

    lines_to_show = [
        lines[i]
        for i in range(
            max(0, line_idx - (MAXIMUM_NUMBER_OF_ERROR_LINES_TO_SHOW - 1)),
            line_idx + 1,
        )
    ]

    return [
        *lines_to_show,
        " " * pos + "^",
    ]



from dataclasses import dataclass
from enum import Enum


class InternalRuleType(Enum):
    CHAR = "CHAR"
    CHAR_RNG_UPPER = "CHAR_RNG_UPPER"
    RULE_REF = "RULE_REF"
    ALT = "ALT"
    END = "END"
    CHAR_NOT = "CHAR_NOT"
    CHAR_ALT = "CHAR_ALT"


@dataclass
class InternalRuleDefWithNumericValue:
    type: InternalRuleType
    value: int


@dataclass
class InternalRuleDefChar:
    type: InternalRuleType
    value: list[int]


@dataclass
class InternalRuleDefCharNot:
    type: InternalRuleType
    value: list[int]


@dataclass
class InternalRuleDefAltChar:
    type: InternalRuleType
    value: int


@dataclass
class InternalRuleDefReference:
    type: InternalRuleType
    value: int


@dataclass
class InternalRuleDefEnd:
    type: InternalRuleType


@dataclass
class InternalRuleDefWithoutValue:
    type: InternalRuleType


InternalRuleDef = (
    InternalRuleDefChar
    | InternalRuleDefCharNot
    | InternalRuleDefWithNumericValue
    | InternalRuleDefWithoutValue
)
InternalRuleDefCharOrAltChar = InternalRuleDefChar | InternalRuleDefAltChar

SymbolIds = dict[str, int]


# Type Guards (Python version)
def is_rule_def_type(type_):
    return isinstance(type_, InternalRuleType)


def is_rule_def(rule):
    return hasattr(rule, "type") and is_rule_def_type(rule.type)


def is_rule_def_alt(rule):
    return getattr(rule, "type", None) == InternalRuleType.ALT


def is_rule_def_ref(rule):
    return getattr(rule, "type", None) == InternalRuleType.RULE_REF


def is_rule_def_end(rule):
    return getattr(rule, "type", None) == InternalRuleType.END


def is_rule_def_char(rule):
    return getattr(rule, "type", None) == InternalRuleType.CHAR


def is_rule_def_char_not(rule):
    return getattr(rule, "type", None) == InternalRuleType.CHAR_NOT


def is_rule_def_char_alt(rule):
    return getattr(rule, "type", None) == InternalRuleType.CHAR_ALT


def is_rule_def_char_rng_upper(rule):
    return getattr(
        rule,
        "type",
        None,
    ) == InternalRuleType.CHAR_RNG_UPPER and isinstance(
        getattr(rule, "value", None),
        int,
    )



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .grammar_graph_types import (
        Pointers,
        ResolvedGraphPointer,
        UnresolvedRule,
        ValidInput,
    )


from ..rules_builder.errors import InputParseError
from ..utils.is_point_in_range import is_point_in_range
from .colorize import colorize
from .get_input_as_code_points import get_input_as_code_points
from .get_serialized_rule_key import get_serialized_rule_key
from .graph_node import GraphNode
from .graph_pointer import GraphPointer
from .rule_ref import RuleRef
from .type_guards import (
    is_range,
    is_rule_char,
    is_rule_char_exclude,
    is_rule_end,
    is_rule_ref,
)

RootNode = dict[int, GraphNode]


def make_pointers() -> Pointers:
    """
    Create a set of pointers with a custom comparison based on pointer ID.
    """
    return set()


class Graph:
    roots: dict[int, RootNode]
    grammar: str
    __rootNode__: RootNode | None
    previous_code_points: list[int]

    def __init__(
        self,
        grammar: str,
        stackedRules: list[list[list[UnresolvedRule]]],
        rootId: int,
    ):
        self.roots = {}
        self.grammar = grammar
        self.previous_code_points = []
        rule_refs: list[RuleRef] = []
        unique_rules: dict[str, UnresolvedRule] = {}

        for stackId in range(len(stackedRules)):
            stack = stackedRules[stackId]
            nodes: dict[int, GraphNode] = {}
            for pathId in range(len(stack)):
                path = stack[pathId]
                node: GraphNode | None = None
                for stepId in range(len(path) - 1, -1, -1):
                    next_node: GraphNode | None = node
                    rule = stack[pathId][stepId]
                    unique_rules[get_serialized_rule_key(rule)] = rule
                    if is_rule_ref(rule):
                        rule_refs.append(rule)
                    # rules coming in may be identical but have different references.
                    # here, we ensure we always use the same reference for an identical rule.
                    # this makes future comparisons easier.
                    unique_rule = unique_rules.get(get_serialized_rule_key(rule))
                    if unique_rule is None:
                        raise ValueError("Could not get unique rule")
                    node = GraphNode(
                        unique_rule,
                        {
                            "stackId": stackId,
                            "pathId": pathId,
                            "stepId": stepId,
                        },
                        next_node,
                    )

                if node is None:
                    raise ValueError("Could not get node")
                nodes[pathId] = node
            self.roots[stackId] = nodes

        self.root_node = self.roots.get(rootId)

        for rule_ref in rule_refs:
            referenced_nodes = set()
            referenced_nodes.update(self.get_root_node(rule_ref.value).values())
            rule_ref.nodes = referenced_nodes

    def get_root_node(self, value: int) -> RootNode:
        root_node = self.roots.get(value)
        if root_node is None:
            raise ValueError(f"Root node not found for value: {value}")
        return root_node

    @property
    def root_node(self) -> RootNode:
        if self.__rootNode__ is None:
            raise ValueError("Root node is not defined")
        return self.__rootNode__

    @root_node.setter
    def root_node(self, rootNode: RootNode | None):
        if rootNode is None:
            raise ValueError("Root node is not defined")
        self.__rootNode__ = rootNode

    def get_initial_pointers(self) -> Pointers:
        pointers = make_pointers()

        for node, parent in self.fetch_nodes_for_root_node(self.root_node):
            pointer = GraphPointer(node, parent)
            for resolvedPointer in self.resolve_pointer(pointer):
                pointers.add(resolvedPointer)
        return pointers

    def set_valid(self, pointers: Pointers, valid: bool):
        for pointer in pointers:
            pointer.valid = valid

    def parse(self, current_pointers: Pointers, codePoint: int) -> Pointers:
        for rule, rule_pointers in self.iterate_over_pointers(current_pointers):
            if is_rule_char(rule):
                valid = False
                for possible_code_point in rule.value:
                    if valid is True:
                        pass
                    elif is_range(possible_code_point):
                        if is_point_in_range(codePoint, possible_code_point):
                            valid = True
                        else:
                            valid = False
                    elif codePoint == possible_code_point:
                        valid = True
                self.set_valid(rule_pointers, valid)

            elif is_rule_char_exclude(rule):
                valid = True
                for possible_code_point in rule.value:
                    if valid is False:
                        pass
                    elif is_range(possible_code_point):
                        if is_point_in_range(codePoint, possible_code_point):
                            valid = False
                        else:
                            valid = True

                self.set_valid(rule_pointers, valid)

            elif not is_rule_end(rule):
                raise ValueError(f"Unsupported rule: {rule}")

        # a pointer's id is the sum of its node's id and its parent's id chain.
        # if two pointers share the same id, it means they point to the same node and have identical parent chains.
        # for the purposes of walking the graph, we only need to keep one of them.
        next_pointers = make_pointers()
        for current_pointer in current_pointers:
            for unresolved_next_pointer in current_pointer.fetch_next():
                for resolved_next_pointer in self.resolve_pointer(
                    unresolved_next_pointer,
                ):
                    next_pointers.add(resolved_next_pointer)
        return next_pointers

    def resolve_pointer(
        self,
        unresolved_pointer: GraphPointer,
    ) -> Iterable[ResolvedGraphPointer]:
        for resolved_pointer in unresolved_pointer.resolve():
            if is_rule_ref(resolved_pointer.node.rule):
                raise ValueError(
                    "Encountered a reference rule when building pointers to the graph",
                )
            if (
                is_rule_end(resolved_pointer.node.rule)
                and resolved_pointer.parent is not None
            ):
                raise ValueError(
                    "Encountered an ending rule with a parent when building pointers to the graph",
                )
            yield resolved_pointer

    def add(self, src: ValidInput, _pointers: Pointers | None = None) -> Pointers:
        pointers = _pointers or self.get_initial_pointers()

        code_points = get_input_as_code_points(src)

        for code_point_pos in range(len(code_points)):
            code_point = code_points[code_point_pos]
            pointers = self.parse(pointers, code_point)
            if len(pointers) == 0:
                raise InputParseError(
                    code_points,
                    code_point_pos,
                    self.previous_code_points,
                )
        self.previous_code_points.extend(code_points)
        return pointers

    #   // generator that yields either the node, or if a reference rule, the referenced node
    #   // we need these function, as distinct from leveraging the logic in GraphPointer,
    #   // because that needs a rule ref with already defined nodes; this function is used to _set_ those nodes
    def fetch_nodes_for_root_node(
        self,
        root_nodes: dict[int, GraphNode],
        parent: GraphPointer | None = None,
    ) -> Iterable[tuple[GraphNode, GraphPointer | None]]:
        for node in root_nodes.values():
            if is_rule_ref(node.rule):
                yield from self.fetch_nodes_for_root_node(
                    self.get_root_node(node.rule.value),
                    GraphPointer(node, parent),
                )
            else:
                yield node, parent

    def print(self, pointers: Pointers | None = None, colors: bool = False) -> str:
        nodes: list[list[GraphNode]] = [
            list(root_node.values()) for root_node in self.roots.values()
        ]
        graph_view: list[str] = []
        for root_node in nodes:
            for node in root_node:
                graph_view.extend(
                    [
                        node.print(
                            {
                                "pointers": pointers or set(),
                                "show_position": True,
                                "colorize": colorize if colors else lambda s, _: str(s),
                            },
                        ),
                    ],
                )

        return "\n".join(graph_view)

    def iterate_over_pointers(
        self,
        pointers: Pointers,
    ) -> Iterable[tuple[UnresolvedRule, Pointers]]:
        seen_rules: dict[UnresolvedRule, Pointers] = {}
        for pointer in pointers:
            rule = pointer.rule
            if is_rule_ref(rule):
                raise ValueError("Encountered a reference rule in the graph")
            seen_rule = seen_rules.get(rule)
            if seen_rule is None:
                seen_rule = set([pointer])
                seen_rules[rule] = seen_rule
            else:
                seen_rule.add(pointer)
        yield from seen_rules.items()



MAXIMUM_NUMBER_OF_ERROR_LINES_TO_SHOW = 3


def GRAMMAR_PARSER_ERROR_HEADER_MESSAGE(reason: str) -> str:
    return f"Failed to parse grammar: {reason}"


INPUT_PARSER_ERROR_HEADER_MESSAGE = "Failed to parse input string:"


class GrammarParseError(Exception):
    def __init__(self, grammar: str, pos: int, reason: str):
        msg = "\n".join(
            [
                GRAMMAR_PARSER_ERROR_HEADER_MESSAGE(reason),
                *build_error_position(grammar, pos),
            ],
        )
        super().__init__(msg)
        self.grammar = grammar
        self.pos = pos
        self.reason = reason
        self.name = "GrammarParseError"


ValidInput = str | int | list[int]


class InputParseError(Exception):
    def __init__(
        self,
        most_recent_input: ValidInput,
        pos: int,
        _previous_input: ValidInput = "",
    ):
        msg = "\n".join(
            [
                INPUT_PARSER_ERROR_HEADER_MESSAGE,
                *build_error_position(get_input_as_string(most_recent_input), pos),
            ],
        )
        super().__init__(msg)
        self.src = most_recent_input
        self.pos = pos
        self.name = "InputParseError"


def build_error_position(src: str, pos: int) -> list[str]:
    if src == "":
        return ["No input provided"]
    grammar_lines = src.split("\n")
    line_idx = 0
    while line_idx < len(grammar_lines) and pos > len(grammar_lines[line_idx]) - 1:
        pos -= len(grammar_lines[line_idx])
        line_idx += 1

    start_line = max(0, line_idx - (MAXIMUM_NUMBER_OF_ERROR_LINES_TO_SHOW - 1))
    lines_to_show = [
        grammar_lines[i]
        for i in range(start_line, line_idx + 1)
        if i < len(grammar_lines)
    ]

    # Append the position marker
    return [*lines_to_show, " " * pos + "^"]


def get_input_as_string(src: ValidInput) -> str:
    if isinstance(src, str):
        return src
    if isinstance(src, list):
        return "".join(chr(cp) for cp in src)
    return str(chr(src))



import pytest

from .get_input_as_string import get_input_as_string


def describe_get_input_as_string():
    @pytest.mark.parametrize(
        ("input", "expected"),
        [
            ("hello", "hello"),
            ([104, 101, 108, 108, 111], "hello"),
            (104, "h"),
            (0x1F600, "😀"),
            ([0x1F600, 0x1F601], "😀😁"),
        ],
    )
    def it_correctly_converts_strings_to_strings(input, expected):
        assert get_input_as_string(input) == expected



from __future__ import annotations

import json

from .grammar_graph_types import (
    RuleChar,
    RuleCharExclude,
    RuleRef,
    RuleType,
    UnresolvedRule,
)
from .type_guards import is_rule_char, is_rule_char_exclude, is_rule_end, is_rule_ref

KEY_TRANSLATION = {
    RuleType.END: 0,
    RuleType.CHAR: 1,
    RuleType.CHAR_EXCLUDE: 2,
}


def get_serialized_rule_key(rule: UnresolvedRule) -> str:
    if is_rule_end(rule):
        return f"{KEY_TRANSLATION[RuleType.END]}"

    if is_rule_char(rule):
        assert isinstance(rule, RuleChar)
        return f"{KEY_TRANSLATION[RuleType.CHAR]}-{json.dumps(rule.value)}"

    if is_rule_char_exclude(rule):
        assert isinstance(rule, RuleCharExclude)
        return f"{KEY_TRANSLATION[RuleType.CHAR_EXCLUDE]}-{json.dumps(rule.value)}"

    if is_rule_ref(rule):
        assert isinstance(rule, RuleRef)
        return f"3-{rule.value}"

    raise ValueError(f"Unknown rule type: {rule}")
